import 'dart:async';

import 'dart:io';
import 'exceptions.dart';
import 'package:quiver/pattern.dart' as quiver;

import 'typedefs.dart';

mixin HttpRequestExecutor {
  FutureOr<void> _executeRoute(
      HttpRequest request, Route route, Context context) async {
    switch (request.method) {
      case 'GET':
        await route.get(request, context);
        break;
      case 'UPDATE':
        await route.update(request, context);
        break;
      case 'POST':
        await route.post(request, context);
        break;
      case 'DELETE':
        await route.delete(request, context);
        break;
      case 'PUSH':
        await route.push(request, context);
        break;
      case 'CONNECT':
        await route.connect(request, context);
        break;
      case 'OPTIONS':
        await route.options(request, context);
        break;
      case 'TRACE':
        await route.trace(request, context);
        break;
      case 'PATCH':
        await route.patch(request, context);
        break;
      default:
        throw IllegalMethodException(request.method);
    }
  }
}

/// Creates a new [Handler] that executes [Middleware] before or after the route's [handler].
/// All [Middleware] is executed sequentially and forms a 'pipeline'.
/// Calling [Context.cancel] inside [Middleware] prevents further [Middleware] from being executed.
Handler registerMiddleware(Handler handler,
    {List<Middleware> before = const [], List<Middleware> after = const []}) {
  return (HttpRequest req, Context con) async {
    for (final middle in before) {
      if (con._canceled) return;
      await middle(req, con);
    }
    if (con._canceled) return;
    await handler(req, con);
    for (final middle in after) {
      if (con._canceled) return;
      await middle(req, con);
    }
  };
}

/// Context for the current route pipeline.
/// The [Context] gets passed from [Middleware] to [Middleware].
/// Anything that gets passed to [values] will be visibel to the other [Middleware]s.
/// Calling [cancel] prevents any further [Middleware] from executing.
class Context {
  var _canceled = false;
  final values = <String, dynamic>{};
  Context._();
  void cancel() {
    _canceled = true;
  }
}

class Route {
  final quiver.Glob route;
  final HttpRequestExecutor executor;
  final Handler get, delete, update, push, post, connect, options, trace, patch;

  Route(String route, this.executor,
      {this.get,
      this.connect,
      this.options,
      this.trace,
      this.patch,
      this.delete,
      this.update,
      this.post,
      this.push})
      : route = quiver.Glob(route);
}

/// Matches the [Uri] and method of [HttpRequest] against registered routes and
/// executes their [Handler]
///
/// Caution: routes get matched in the order they are registered.
/// If you use globs, register routes from specific to less specific.
class HttpRouter with HttpRequestExecutor {
  static FutureOr<void> _standardNotFoundHandler(
      HttpRequest req, Context context) {
    req.response.statusCode = HttpStatus.notFound;
    req.response.writeln(req.response.reasonPhrase);
    req.response.close();
  }

  final _routes = <Route>[];
  final Handler _notFound;

  /// [notFoundHandler] gets executed when no registered route matches
  /// the [HttpRequest]
  HttpRouter({Handler notFoundHandler = _standardNotFoundHandler})
      : _notFound = notFoundHandler;

  void _checkAlreadeRegisteredRoute(String route) {
    final alreadyRegisteredRoute =
        _routes.firstWhere((r) => r.route.pattern == route, orElse: () => null);
    if (alreadyRegisteredRoute != null) {
      throw StateError('Route "$route" has already been registered');
    }
  }

  /// Registeres a [route] and specific [Handler]s for specific Http methods.
  void register(String route,
      {Handler get,
      Handler delete,
      Handler update,
      Handler post,
      Handler push,
      Handler connect,
      Handler options,
      Handler trace,
      Handler patch}) {
    _checkAlreadeRegisteredRoute(route);
    _routes.add(Route(route, this,
        get: get,
        connect: connect,
        options: options,
        trace: trace,
        patch: patch,
        delete: delete,
        update: update,
        post: post,
        push: push));
  }

  /// Matches the [request] against registered routes and
  /// executes their [Handler]
  FutureOr<void> handleRequest(HttpRequest request) async {
    final route = _routes.firstWhere(
        (route) => route.route.matchAsPrefix(request.uri.path) != null,
        orElse: () => null);

    final context = Context._();
    if (route == null) {
      _notFound(request, context);
      return;
    }
    await route.executor._executeRoute(request, route, context);
  }

  HttpChildRouter child(
      {List<Middleware> before = const [], List<Middleware> after = const []}) {
    final child = HttpChildRouter._(this, before, after);
    return child;
  }
}

class HttpChildRouter with HttpRequestExecutor {
  final HttpRouter _parentRouter;
  final List<Middleware> _before, _after;
  HttpChildRouter._(this._parentRouter, this._before, this._after);

  /// Registeres a [route] and specific [Handler]s for specific Http methods.
  void register(String route,
      {Handler get,
      Handler delete,
      Handler update,
      Handler post,
      Handler push,
      Handler connect,
      Handler options,
      Handler trace,
      Handler patch}) {
    _parentRouter._checkAlreadeRegisteredRoute(route);
    _parentRouter._routes.add(Route(route, this,
        get: get,
        connect: connect,
        options: options,
        trace: trace,
        patch: patch,
        delete: delete,
        update: update,
        post: post,
        push: push));
  }

  @override
  FutureOr<void> _executeRoute(
      HttpRequest request, Route route, Context context) async {
    for (final middle in _before) {
      if (context._canceled) return;
      await middle(request, context);
    }
    if (context._canceled) return;
    super._executeRoute(request, route, context);
    for (final middle in _after) {
      if (context._canceled) return;
      await middle(request, context);
    }
  }
}
